from handshake.services.DBService.models.result_base import LogType, TestLogBase
from handshake.services.DBService.shared import get_test_id
from sanic.request import Request
from sanic.response import JSONResponse
from pydantic import ValidationError


def extractPayload(request: Request, response: JSONResponse):
    payload = dict(
        url=request.url,
        payload=request.json,
        status=response.status,
        reason=response.body.decode(),
    )
    return payload


def extractPydanticErrors(url, payload, error: ValidationError):
    payload = dict(
        url=url,
        payload=payload,
        status="ERROR",
        reason=error.json(),
    )
    return payload


async def attachLog(payload, attachmentType: LogType, description: str):
    await TestLogBase.create(
        test_id=get_test_id(),
        type=attachmentType,
        feed=payload,
        message=description,
    )


async def attachError(payload, url: str):
    await attachLog(
        payload,
        LogType.ERROR,
        f"Failed to process the request at: {url}, will affect the test run",
    )


async def attachInfo(payload, url: str):
    await attachLog(
        payload,
        LogType.INFO,
        f"utilizing: {url}",
    )


async def attachWarn(payload, url: str):
    await attachLog(
        payload,
        LogType.WARN,
        f"Failed to process the request at: {url}, we will miss this attachment",
    )



import json
from handshake.services.DBService.models.config_base import ConfigBase
from handshake.services.DBService import DB_VERSION
from handshake.services.DBService.models.result_base import RunBase
from handshake.services.DBService.models.enums import ConfigKeys
from handshake.services.DBService.migrator import migration
from tortoise import Tortoise, connections
from handshake.services.DBService.shared import db_path
from pathlib import Path
from typing import Optional, Union
from loguru import logger
from handshake.services.SchedularService.constants import (
    writtenAttachmentFolderName,
)

models = ["handshake.services.DBService.models"]


def config_file(provided_db_path: Path):
    return provided_db_path.parent / "config.json"


def attachment_folder(provided_db_path: Path):
    return provided_db_path.parent / writtenAttachmentFolderName


async def init_tortoise_orm(
    force_db_path: Optional[Union[Path, str]] = None, migrate: bool = False
):
    chosen = force_db_path if force_db_path else db_path()
    if migrate:
        migration(chosen)

    await Tortoise.init(
        db_url=r"{}".format(f"sqlite://{chosen}"),
        modules={"models": models},
    )
    await Tortoise.generate_schemas()
    await set_default_config(chosen)


async def create_run(projectName: str) -> str:
    test_id = str((await RunBase.create(projectName=projectName)).testID)
    return test_id


READ_ONLY = (
    ConfigKeys.version,
    ConfigKeys.recentlyDeleted,
    ConfigKeys.reset_test_run,
)
ALLOW_WRITE = {
    ConfigKeys.maxRunsPerProject,
}


async def set_default_config(path: Path):
    attachment_folder(path).mkdir(exist_ok=True)
    config_file_provided = config_file(path)
    config_provided = (
        json.loads(config_file_provided.read_text())
        if config_file_provided.exists()
        else dict()
    )

    for key, value in [
        (ConfigKeys.version, DB_VERSION),
        (ConfigKeys.reset_test_run, ""),
        (ConfigKeys.maxRunsPerProject, "100")
        # below keys can be overridden by the config file
    ]:
        record = await ConfigBase.filter(key=str(key)).first()
        if not record:
            logger.debug(
                "{} was not found in our table, registering it with value: {}",
                key,
                value,
            )
            await ConfigBase.create(key=key, value=value)
        else:
            if (
                record.key in ALLOW_WRITE
                and config_provided.get(record.key, value) != record.value
            ):
                logger.debug(
                    "Found a key: {} already existing in configbase with value: {},"
                    " but received a request to change it to {}",
                    record.key,
                    value,
                    config_provided.get(record.key, value),
                )
                record.value = config_provided.get(record.key, value)
                await record.save()


async def close_connection():
    await connections.close_all()
    # waiting for the logs to be sent or saved
    await logger.complete()



import json
from asyncio import gather, to_thread, TaskGroup
from handshake.services.DBService.lifecycle import (
    init_tortoise_orm,
    db_path,
    close_connection,
)
from shutil import rmtree
from tarfile import open
from pathlib import Path
from typing import Optional
from loguru import logger
from tortoise.expressions import Q, RawSQL
from handshake.services.DBService.models.result_base import (
    RunBase,
    SuiteBase,
    SuiteType,
    SessionBase,
)
from tortoise import connections, BaseDBAsyncClient
from handshake.services.DBService.models.config_base import (
    ConfigBase,
    ConfigKeys,
)
from handshake.services.SchedularService.constants import (
    writtenAttachmentFolderName,
    exportAttachmentFolderName,
    EXPORT_RUN_PAGE_FILE_NAME,
    EXPORT_PROJECTS_FILE_NAME,
    EXPORT_RUNS_PAGE_FILE_NAME,
    EXPORT_OVERVIEW_PAGE,
    EXPORT_ALL_SUITES,
)
from handshake.services.DBService.models.dynamic_base import TaskBase, JobType
from handshake.services.SchedularService.pruneTasks import pruneTasks
from handshake.services.SchedularService.handleTestResults import delete_test_attachment
from handshake.services.SchedularService.handlePending import patch_jobs
from handshake.services.SchedularService.refer_types import (
    SubSetOfRunBaseRequiredForProjectExport,
    SuiteSummary,
)


class Scheduler:
    RUNS_PAGE_VERSION = 1

    def __init__(
        self,
        root_dir: str,
        out_dir: Optional[str] = None,
        manual_reset: Optional[bool] = False,
        zipped_build: Optional[str] = None,
        include_build: Optional[bool] = False,
    ):
        self.export = out_dir is None
        self.dashboard_build = zipped_build
        self.export_dir = Path(out_dir) if out_dir and zipped_build else None
        self.db_path = db_path(root_dir)
        self.include_build = include_build
        self.import_dir = (
            self.db_path.parent / exportAttachmentFolderName
            if not self.export_dir
            else self.export_dir / exportAttachmentFolderName
        )
        self.reset = manual_reset
        self.connection: Optional[BaseDBAsyncClient] = None

    async def start(self):
        await init_tortoise_orm(self.db_path, True)
        self.connection = connections.get("default")
        await self.rotate_test_runs()
        await self.init_jobs()
        await patch_jobs()
        await self.export_jobs()

        await close_connection()

    async def rotate_test_run(self, projectName: str, to_delete: int):
        project_runs = RunBase.filter(projectName=projectName).all()
        will_be_deleted = await project_runs.count() - to_delete

        if will_be_deleted <= 0:
            logger.debug(
                "Test Runs for the project: {} has not yet reached the limit,"
                " hence skipping the deletion Task for this project",
                projectName,
            )
            return 0

        to_delete_runs = await (
            project_runs.order_by("started")
            .limit(will_be_deleted)
            .values_list("testID", flat=True)
        )
        attachment_folder = self.db_path.parent / writtenAttachmentFolderName
        await gather(
            RunBase.filter(Q(testID__in=to_delete_runs)).delete(),
            *(
                to_thread(delete_test_attachment, test_id, attachment_folder)
                for test_id in to_delete_runs
            ),
        )

        logger.warning(
            "Deleted {} test runs of this project {}", will_be_deleted, projectName
        )
        return will_be_deleted

    async def rotate_test_runs(self):
        runs_per_project = await ConfigBase.filter(
            key=ConfigKeys.maxRunsPerProject
        ).first()
        requested = max(int(runs_per_project.value), 2)

        recently_deleted = sum(
            await gather(
                *(
                    self.rotate_test_run(projectName, requested)
                    for projectName in await RunBase.all()
                    .distinct()
                    .values_list("projectName", flat=True)
                )
            )
        )

        record = await ConfigBase.filter(key=ConfigKeys.recentlyDeleted).first()
        if record:
            await record.update_from_dict(dict(value=str(recently_deleted)))
            await record.save()
        else:
            await ConfigBase.create(
                key=ConfigKeys.recentlyDeleted, value=str(recently_deleted)
            )

        logger.info("Delete job is completed.")

    async def init_jobs(self):
        await pruneTasks()

        prev_picked_tasks = await TaskBase.filter(processed=False, picked=True).all()
        reset_from_config = await ConfigBase.filter(
            key=ConfigKeys.reset_test_run
        ).first()

        self.reset = self.reset or reset_from_config.value

        if self.reset:
            logger.info("It was requested to reset all of the completed test runs")

        to_modify_test_runs = (
            await TaskBase.filter(type=JobType.MODIFY_TEST_RUN, processed=True).all()
            if self.reset
            else []
        )
        to_pick = prev_picked_tasks + to_modify_test_runs

        async def pick_old_tasks():
            for task in to_pick:
                logger.info("scheduling old task {} for this iteration", task.ticketID)
                task.picked = False
                task.processed = False

            if to_pick:
                await TaskBase.bulk_update(to_pick, ("picked", "processed"), 100)
                logger.debug("Done!, Marked {} old tasks for processing", len(to_pick))

        async def reset_completed_runs():
            runs = []
            for ticket in to_modify_test_runs:
                test: RunBase = await ticket.test
                logger.debug("Reset for test run: {}", test.testID)
                test.standing = "PENDING"
                runs.append(test)

            if runs:
                logger.info("Reset done for {} test runs!", len(runs))
                await RunBase.bulk_update(runs, ("standing",), 100)

            if reset_from_config:
                reset_from_config.value = ""
                await reset_from_config.save()

        await gather(pick_old_tasks(), reset_completed_runs())
        logger.debug("Pre-Patch Jobs have been initiated")

    async def export_jobs(self):
        if not self.export_dir:
            logger.debug("Skipping export, as the output directory was not provided.")
            if not self.include_build:
                return

        if self.export_dir:
            logger.debug("Exporting reports...")
            await to_thread(self.export_files)
            logger.info("Exporting results...")

        self.import_dir.mkdir(exist_ok=self.include_build)

        await self.export_runs_page()
        logger.info("Done!")

    async def export_runs_page(self):
        logger.debug("Exporting Runs Page...")

        async with TaskGroup() as exporter:
            runs = []
            projects = {}

            for row in (
                await self.connection.execute_query(
                    """
select rb.*, cb.*,
rank() over (order by rb.ended desc) as timelineIndex,
rank() over (partition by projectName order by rb.ended desc) as projectIndex
from RUNBASE rb
left join testconfigbase cb on rb.testID = cb.test_id 
WHERE rb.ended <> '' order by rb.started;
-- note for: projectIndex and timelineIndex, latest -> oldest => 0 - ...
"""
                )
            )[-1]:
                run = dict(row)
                test_run = SubSetOfRunBaseRequiredForProjectExport.model_validate(run)
                runs.append(run)

                logger.info(
                    "Exporting runs page for {} - {}",
                    test_run.projectName,
                    test_run.testID,
                )

                projects[test_run.projectName] = projects.get(test_run.projectName, [])
                suite_summary: SuiteSummary = json.loads(test_run.suiteSummary)
                projects[test_run.projectName].append(
                    dict(
                        testID=test_run.testID,
                        passed=test_run.passed,
                        failed=test_run.failed,
                        skipped=test_run.skipped,
                        tests=test_run.tests,
                        passedSuites=suite_summary["passed"],
                        failedSuites=suite_summary["failed"],
                        skippedSuites=suite_summary["skipped"],
                        suites=suite_summary["count"],
                        duration=test_run.duration,
                    )
                )

                (self.import_dir / str(test_run.testID)).mkdir(exist_ok=True)

                exporter.create_task(
                    to_thread(
                        self.save_run_query,
                        str(test_run.testID),
                        json.dumps(run),
                    ),
                    name="export-run-page",
                )

                exporter.create_task(
                    self.export_run_page(test_run.testID),
                    name="export-more-for-run-page",
                )

            exporter.create_task(
                to_thread(self.save_runs_query, json.dumps(runs), json.dumps(projects)),
                name="export-runs-page",
            )

            logger.info("Exported Runs Page!")

    async def export_run_page(self, run_id: str):
        await gather(self.export_overview_page(run_id), self.export_all_suites(run_id))

    async def export_overview_page(self, run_id: str):
        recent_suites = await (
            SuiteBase.filter(Q(session__test_id=run_id) & Q(suiteType=SuiteType.SUITE))
            .order_by("-started")
            .limit(6)
            .annotate(
                # numberOfErrors=RawSQL("json_array_length(errors)"),
                id=RawSQL("suiteID"),
                s=RawSQL("suitebase.started"),
            )
            .values(
                "title",
                "passed",
                "failed",
                "skipped",
                "duration",
                suiteID="id",
                started="s",
            )
        )

        aggregated = (
            await SuiteBase.filter(session__test_id=run_id)
            .annotate(
                sessions=RawSQL("count(DISTINCT session_id)"),
                files=RawSQL("count(DISTINCT file)"),
            )
            .only("sessions", "files")
            .first()
            .values("sessions", "files")
        )

        platforms = (
            await SessionBase.filter(test_id=run_id)
            .only("entityName", "entityVersion", "simplified")
            .distinct()
            .values("entityName", "entityVersion", "simplified")
        )

        await to_thread(
            self.save_overview_query,
            run_id,
            json.dumps(
                dict(
                    recentSuites=recent_suites,
                    aggregated=aggregated,
                    platforms=platforms,
                )
            ),
        ),

    async def export_all_suites(self, run_id: str):
        all_suites = await (
            SuiteBase.filter(Q(session__test_id=run_id) & Q(suiteType=SuiteType.SUITE))
            .order_by("started")
            .prefetch_related("rollup")
            .annotate(
                numberOfErrors=RawSQL("json_array_length(errors)"),
                id=RawSQL("suiteID"),
                p_id=RawSQL("parent"),
                s=RawSQL("suitebase.started"),
                e=RawSQL("suitebase.ended"),
                error=RawSQL("errors ->> '[0]'"),
                hasChildSuite=RawSQL(
                    "(select count(*) from suitebase sb where sb.parent=suitebase.suiteID "
                    "and sb.suiteType='SUITE' LIMIT 1)"
                ),
            )
            .values(
                "title",
                "passed",
                "failed",
                "standing",
                "skipped",
                "duration",
                "file",
                "retried",
                "tags",
                "description",
                "errors",
                "error",
                "numberOfErrors",
                "hasChildSuite",
                suiteID="id",
                parent="p_id",
                started="s",
                ended="e",
                entityName="session__entityName",
                entityVersion="session__entityVersion",
                hooks="session__hooks",
                simplified="session__simplified",
                rollup_passed="rollup__passed",
                rollup_failed="rollup__failed",
                rollup_skipped="rollup__skipped",
                rollup_tests="rollup__tests",
            )
        )

        await to_thread(
            self.save_all_suites_query,
            run_id,
            json.dumps(all_suites),
        )

    def export_files(self):
        # we reset entire export folder
        if self.export_dir.exists():
            logger.debug("removing previous results")
            rmtree(self.export_dir)

        self.export_dir.mkdir(exist_ok=False)

        with open(self.dashboard_build, "r:bz2") as tar_file:
            tar_file.extractall(self.export_dir)

    def save_runs_query(self, feed: str, projectsFeed: str):
        (self.import_dir / EXPORT_RUNS_PAGE_FILE_NAME).write_text(feed)
        (self.import_dir / EXPORT_PROJECTS_FILE_NAME).write_text(projectsFeed)

    def save_run_query(self, testID: str, feed: str):
        (self.import_dir / testID / EXPORT_RUN_PAGE_FILE_NAME).write_text(feed)

    def save_overview_query(self, testID: str, feed: str):
        (self.import_dir / testID / EXPORT_OVERVIEW_PAGE).write_text(feed)

    def save_all_suites_query(self, testID: str, feed: str):
        (self.import_dir / testID / EXPORT_ALL_SUITES).write_text(feed)



from handshake.services.DBService.models.enums import ConfigKeys
from tortoise.fields import IntField
from handshake.services.DBService.models.result_base import RunBase
from tortoise.models import Model
from tortoise.fields import (
    CharEnumField,
    ForeignKeyField,
    ForeignKeyRelation,
    TextField,
    UUIDField,
    BooleanField,
    JSONField,
)


class ConfigBase(Model):
    key = CharEnumField(
        ConfigKeys, pk=True, null=False, description="Type of job we would like to run"
    )
    value = TextField(null=False, description="Handling type is up to us")


class ExportBase(Model):
    ticketID = UUIDField(pk=True)
    maxTestRuns = IntField(
        null=True,
        default=10,
        description="Number of test runs to export [recent ones are picked]",
    )


class TestConfigBase(Model):
    test: ForeignKeyRelation[RunBase] = ForeignKeyField(
        "models.RunBase", related_name="runs", to_field="testID"
    )
    platform = TextField(null=False, description="could be windows or linux")
    framework = TextField(null=False, description="name of the framework used")
    maxInstances = IntField(
        null=True, default=1, description="Max. Number of workers used to run the tests"
    )
    fileRetries = IntField(
        null=False,
        default=0,
        description="Number of times it can retry a spec file",
    )
    avoidParentSuitesInCount = BooleanField(
        null=False,
        default=False,
        description="whether to count the spec file as a suite or not,"
        " example: in gherkin if in a feature file with one scenario, "
        "if this value is true then the number of suites is treated as 1 else 2",
    )
    bail = IntField(
        null=False,
        default=0,
        description="if > 0 then it means that run would stop if it finds this number of test cases failed",
    )
    tags = JSONField(
        default=[],
        null=False,
        description="comma separated list of tags (used by framework) to filter the spec files",
    )



from pydantic import ValidationError
from sanic.request import Request
from sanic.response import JSONResponse, json


async def handle_validation_error(request: Request, validation_error: ValidationError) -> JSONResponse:
    return json(validation_error.errors(), status=400)



import os
from handshake.services.DBService.lifecycle import DB_VERSION
from handshake.services.Endpoints.internalEndpoints import one_liners
from handshake.services.DBService.shared import APP_NAME
from handshake.services.Endpoints.blueprints.coreEndpoints import service
from handshake.services.Endpoints.blueprints.writeServices import writeServices
from handshake.services.Endpoints.errorHandling import handle_validation_error
from sanic import Sanic
from pydantic import ValidationError
import sentry_sdk
from handshake import __version__

dns = os.getenv("SENTRY_SDK")
if dns:
    sentry_sdk.init(
        dsn=os.getenv("SENTRY_SDK"),
        enable_tracing=True,
        max_breadcrumbs=10,
        release=f"{__version__} - {DB_VERSION}",
        server_name=APP_NAME,
    )

service_provider = Sanic(APP_NAME)
service_provider.config.TOUCHUP = False
service_provider.blueprint(one_liners)
service_provider.blueprint(service)
service_provider.blueprint(writeServices)

service_provider.error_handler.add(ValidationError, handle_validation_error)



import typing
from sqlite3 import connect
from handshake.services.DBService import DB_VERSION
from handshake.services.DBService.models.enums import ConfigKeys
from sqlite3.dbapi2 import Connection
from typing import Optional, Tuple
from pathlib import Path
from loguru import logger


def check_version(
    path: Path = None, connection: Optional[Connection] = None
) -> Tuple[bool, Connection, bool, Optional[int]]:
    connection = connection if connection else connect(path)
    query = f"select value from configbase where key = '{ConfigKeys.version}'"
    result = connection.execute(query).fetchone()

    version_stored = False
    migration_required = False

    if not result:
        logger.warning(
            f"Could not find the version, Please raise this as an issue or re-run after deleting the folder at: {path}."
        )
    else:
        version_stored = result if not result else int(result[0])
        migration_required = version_stored != DB_VERSION

        logger.log(
            "INFO" if not migration_required else "WARNING",
            "Currently at: v{}."
            if not migration_required
            else 'Found version: v{}. but required is v{}. Please execute: \n"handshake migrate [COLLECTION_PATH]"',
            result[0],
            DB_VERSION,
        )
    return (
        migration_required,
        connection,
        version_stored < DB_VERSION,
        version_stored,
    )


# returns True if it requires further migration
def migrate(
    connection: Optional[Connection] = None, db_path: Optional[Path] = None
) -> typing.Tuple[bool, bool]:  # further migration required, migration was done
    if not connection:
        connection = connect(db_path)

    try:
        is_required, connection, bump_if_required, version_stored = check_version(
            connection=connection
        )

        if not is_required:
            logger.info("Already migrated to required version")
            return False, False

        if not bump_if_required:
            logger.error(
                "You have more recent version of database, v{}. but we can only support: v{}. "
                "Hence requesting you to update your reporter, to support this version.",
                # "use your backup database inside the collection_path.",
                version_stored,
                DB_VERSION,
            )
            return False, False

        script = Path(__file__).parent / "scripts" / f"bump-v{version_stored}.sql"
        logger.info("Executing {} to migrate from v{}", script.name, version_stored)
        connection.executescript(script.read_text())
    finally:
        if db_path:
            connection.commit()
            connection.close()

    return (version_stored + 1) < DB_VERSION, version_stored < DB_VERSION


def migration(path: Path):
    migration_was_conducted = False
    if not path.exists():
        logger.info("Migration check is not required, as the db does not exist.")
        return

    connection = connect(path)
    try:
        backup_file = "backup_results.db"

        while True:
            first_time = not migration_was_conducted
            further_required, migration_was_conducted = migrate(connection)
            if migration_was_conducted and first_time:
                backup_path = path.parent / backup_file
                backup_path.unlink(missing_ok=True)
                backup = connect(backup_path)
                logger.info(
                    "Before migration, Taking a backup and storing it in {}",
                    backup_path,
                )
                connection.backup(backup)
                backup.close()
            if not migration_was_conducted:
                migration_was_conducted = True
                break

        if migration_was_conducted:
            (path.parent / backup_file).unlink(missing_ok=True)
            connection.commit()
            logger.info("Migrated to latest version!")

    except Exception as error:
        logger.error(f"Failed to execute migration script, due to {error}")
        return True
    finally:
        connection.close()


if __name__ == "__main__":
    migration(Path.cwd().parent.parent.parent / "TestResults" / "TeStReSuLtS.db")



from loguru import logger
from handshake.services.DBService.models.dynamic_base import TaskBase, JobType
from uuid import uuid4, UUID
from typing import Union
from handshake.services.DBService.models.result_base import TestLogBase, LogType
from handshake.services.DBService.lifecycle import init_tortoise_orm, close_connection
from handshake.services.DBService.models.config_base import ExportBase
from handshake.services.DBService.models.result_base import RunBase
from pathlib import Path
from typing import List


async def register_patch_suite(suiteID: str, testID: str) -> TaskBase:
    return await TaskBase.create(
        ticketID=suiteID, test_id=testID, type=JobType.MODIFY_SUITE
    )


async def register_patch_test_run(testID: str) -> TaskBase:
    return await TaskBase.create(
        type=JobType.MODIFY_TEST_RUN, test_id=testID, ticketID=testID
    )


async def mark_for_prune_task(test_id: str):
    # someone called this explicitly hence it's a warning

    logger.warning("Requested to prune some tasks")
    await TaskBase.create(
        ticketID=str(uuid4()), type=JobType.PRUNE_TASKS, test_id=test_id
    )


async def skip_test_run(test_id: Union[str, UUID], reason: str, **extra) -> False:
    logger.error(reason)
    await TestLogBase.create(
        test_id=str(test_id), message=reason, type=LogType.ERROR, feed=extra
    )
    await mark_for_prune_task(test_id)
    return False


async def warn_about_test_run(test_id: Union[str, UUID], about: str, **extra) -> True:
    logger.warning(about)
    await TestLogBase.create(
        test_id=str(test_id), message=about, type=LogType.WARN, feed=extra
    )
    return True



from handshake.services.CommandLine._init import (
    handle_cli,
    general_but_optional_requirement,
    version_option,
    observed_version,
    break_if_mismatch,
)
from handshake.services.Endpoints.center import service_provider
from handshake.services.DBService.lifecycle import (
    init_tortoise_orm,
    close_connection,
    create_run,
)
from handshake import __version__
from typing import Union
from handshake.services.DBService.shared import set_test_id
from click import argument, option, Path
from pathlib import Path as P_Path
from multiprocessing.sharedctypes import Array
from sanic.worker.loader import AppLoader
from sanic import Sanic
from typing import Tuple
from functools import partial
from loguru import logger
from handshake.services.Endpoints.static_server import static_provider


def feed_app() -> Sanic:
    return service_provider


def feed_static_provider(export: str) -> Sanic:
    static_provider.static(
        "/", str(P_Path(export)), name="export", index=["index.html", "RUNS.html"]
    )
    return static_provider


def prepare_loader() -> Tuple[Sanic, AppLoader]:
    loader = AppLoader(factory=feed_app)
    app = loader.load()
    return app, loader


def setup_app(
    projectname: str,
    path: str,
    port: int = 6969,
    workers: int = 2,
):
    @service_provider.main_process_start
    async def get_me_started(app, loop):
        service_provider.shared_ctx.ROOT = Array("c", str.encode(path))
        await init_tortoise_orm(migrate=True)
        test_id = await create_run(projectname)
        service_provider.shared_ctx.TEST_ID = Array("c", str.encode(test_id))
        set_test_id()

    @service_provider.main_process_stop
    async def close_things(app, loop):
        await close_connection()

    _app, loader = prepare_loader()
    _app.prepare(
        port=port,
        workers=max(2, workers),
        host="127.0.0.1",
        motd_display=dict(version=__version__),
    )
    logger.debug("Serving at port: {}", port)
    Sanic.serve(primary=_app, app_loader=loader)


@handle_cli.command(
    short_help="Starts the server which would listen for your input",
    help="""
Starts the Handshake server to listen for inputs at the specified port on localhost. This command initiates a test run,
allowing the server to handle a single test run at a time. For multiple test runs, spawn the process separately on different
ports.
""",
)
@argument("PROJECT_NAME", nargs=1, required=True, type=str)
@observed_version
@general_but_optional_requirement
@option(
    "-p",
    "--port",
    default=6969,
    show_default=True,
    help="Port for the service to connect to",
    type=int,
)
@option(
    "-w",
    "--workers",
    default=2,
    show_default=True,
    help="Number of workers to use",
    type=int,
)
def run_app(
    collection_path: str,
    project_name: str,
    version: Union[str, bool],
    port: int,
    workers: int,
):
    break_if_mismatch(version)
    if workers < 2:
        logger.warning(
            "we have set default of 2 workers, if it's less than that, server might miss results sent from the runner."
        )

    P_Path(collection_path).mkdir(exist_ok=True)
    setup_app(project_name, collection_path, port, workers)


@handle_cli.command(
    help="serves the generated reports. simply serves the static files generated in your directory mentioned "
    "in static_path",
    short_help="serves generated report",
)
@argument("STATIC_PATH", nargs=1, required=False, type=Path(exists=True, dir_okay=True))
@option(
    "-p",
    "--port",
    default=8000,
    show_default=True,
    help="Port for the service to connect to",
    type=int,
)
def display(
    version: Union[bool, str] = False,
    static_path: Union[str, P_Path] = "TestReports",
    port: int = 8000,
):
    if static_path:
        static_path = P_Path(static_path)

        if not static_path.exists():
            raise NotADirectoryError(f"{static_path} does not exist")

    loader = AppLoader(factory=partial(feed_static_provider, static_path))
    _app = loader.load()
    _app.prepare(
        host="127.0.0.1",
        port=port,
        access_log=False,
        motd_display=dict(version=__version__),
    )
    Sanic.serve(primary=_app, app_loader=loader)


if __name__ == "__main__":
    setup_app("sample", "../../../TestResults")



from sanic import Sanic
from pathlib import Path
from handshake.services.DBService.sanic_free_shared import db_name
from typing import Optional, Union

APP_NAME = "Handshake"


def root_dir() -> Path:
    return Path(Sanic.get_app(APP_NAME).shared_ctx.ROOT.value.decode("utf-8"))


def db_path(given_root: Optional[Union[str, Path]] = None) -> Path:
    if not given_root:
        return root_dir() / db_name()

    return Path(given_root) / db_name()


def set_test_id():
    app: Sanic = Sanic.get_app(APP_NAME)
    if not hasattr(app.shared_ctx, "TEST_ID"):
        return
    # ease of access so
    app.config.TEST_ID = app.shared_ctx.TEST_ID.value.decode("utf-8")


def get_test_id() -> str:
    return str(Sanic.get_app(APP_NAME).config.TEST_ID)



from pydantic import BaseModel
from typing import List, Dict, Union, TypedDict, Optional

Node = TypedDict("Node", {"<path>": str, "<count>": Optional[int]}, total=False)


class PathTree(TypedDict):
    current: str
    suites: Optional[int]
    paths: Dict[str, "PathTree"]


class PathItem(TypedDict):
    pointer: PathTree
    children: List[str]
    count: int


class SuiteSummary(TypedDict):
    passed: int
    skipped: int
    count: int
    failed: int


class SubSetOfRunBaseRequiredForProjectExport(BaseModel):
    testID: str
    tests: int
    passed: int
    skipped: int
    failed: int
    suiteSummary: str
    duration: float
    projectName: str



from handshake.services.DBService.models.result_base import SuiteBase
from tortoise.models import Model
from tortoise.fields import (
    ForeignKeyField,
    ForeignKeyRelation,
    TextField,
    IntField,
    BooleanField,
)


class AssertBase(Model):
    entity: ForeignKeyRelation[SuiteBase] = ForeignKeyField(
        "models.SuiteBase", related_name="assertion", to_field="suiteID"
    )
    passed = BooleanField(
        default=False, description="Whether the assertion passed or not"
    )
    wait = IntField(
        description="Number of milli-seconds configured to wait for this test",
        default=-1,
        null=True,
    )
    interval = IntField(
        description="interval (in milli-seconds) to test this assertion until it passes",
        default=-1,
        null=True,
    )
    title = TextField(description="Name of the Assertion")
    message = TextField(description="Message attached to the assertion")



from enum import StrEnum


class JobType(StrEnum):
    INIT_CONNECTION_JOBS = "set-db-connection-add-jobs"
    LOOKUP_JOB = "pending-tasks"
    MODIFY_SUITE = "fix-suite"
    MODIFY_TEST_RUN = "fix-test-run"
    EXECUTOR = "say-bye-if-required"
    PRUNE_TASKS = "prune-tasks"
    DELETE_RUNS = "delete-runs"


writtenAttachmentFolderName = "Attachments"
exportAttachmentFolderName = "Import"

EXPORT_RUNS_PAGE_FILE_NAME = "runs.json"
EXPORT_RUN_PAGE_FILE_NAME = "run.json"
EXPORT_PROJECTS_FILE_NAME = "projects.json"
EXPORT_OVERVIEW_PAGE = "overview.json"
EXPORT_ALL_SUITES = "suites.json"



from tortoise.models import Model
from tortoise.fields import (
    CharField,
    DatetimeField,
    JSONField,
    CharEnumField,
    ForeignKeyField,
    ForeignKeyRelation,
    BooleanField,
    TextField,
)
from handshake.services.SchedularService.constants import JobType
from handshake.services.DBService.models.result_base import RunBase
from handshake.services.DBService.models.enums import PrunedRecords


class TaskBase(Model):
    table = "TaskBase"

    ticketID = CharField(max_length=45, pk=True)
    type = CharEnumField(JobType, null=False)
    dropped = DatetimeField(auto_now=True)  # use modified timestamp
    # this would schedule the parent suites in the later rounds
    meta = JSONField(
        null=True,
        default={},
        description="Data required to process the task, Not used as of now though",
    )
    test: ForeignKeyRelation[RunBase] = ForeignKeyField(
        "models.RunBase", related_name="test", to_field="testID"
    )
    picked = BooleanField(
        null=True,
        default=False,
        description="True if the task is picked by the job else False",
    )
    processed = BooleanField(
        null=True,
        default=False,
        description="True if the task is processed or completed",
    )


class PrunedBase:
    prunedID = CharField(
        max_length=36, description="possibly uuid of length 36 that was pruned"
    )
    reason = TextField(description="Reason for its existence here")
    details = JSONField(
        null=True, default={}, description="If any details were required"
    )
    type = CharEnumField(PrunedRecords, null=False)



from handshake.services.DBService.models.result_base import SessionBase, SuiteBase
from handshake.services.DBService.models.types import (
    RegisterSession,
    RegisterSuite,
    MarkSuite,
    MarkSession,
    AddAttachmentForEntity,
    PydanticModalForTestRunConfigBase,
    WrittenAttachmentForEntity,
)
from handshake.services.Endpoints.blueprints.utils import (
    attachError,
    extractPayload,
    attachWarn,
    extractPydanticErrors,
)
from handshake.services.DBService.models.static_base import (
    AttachmentBase,
    AttachmentType,
)
from handshake.services.DBService.models.config_base import TestConfigBase
from handshake.services.DBService.models.attachmentBase import AssertBase
from handshake.services.DBService.models.enums import Status, SuiteType
from handshake.services.DBService.models.static_base import StaticBase
from sanic.blueprints import Blueprint
from sanic.response import JSONResponse, text, HTTPResponse
from loguru import logger
from sanic.request import Request
from handshake.services.DBService.shared import get_test_id
from handshake.services.SchedularService.register import register_patch_suite
from pydantic import ValidationError
from dotenv import load_dotenv
from handshake.services.DBService.lifecycle import attachment_folder, db_path

load_dotenv()


service = Blueprint("DBService", url_prefix="/save")


@service.on_response
async def handle_response(request: Request, response: JSONResponse):
    if 200 <= response.status < 300:
        return response

    payload = extractPayload(request, response)
    await attachError(payload, request.url)
    return JSONResponse(body=payload, status=response.status)


@service.put("/registerSession")
async def register_session(request: Request) -> HTTPResponse:
    try:
        session = RegisterSession.model_validate(request.json)
        session_record = await SessionBase.create(
            **session.model_dump(), test_id=get_test_id()
        )
        await session_record.save()
    except Exception as error:
        logger.error("Failed to create a session due to exception: {}", str(error))
        return text(str(error), status=404)

    return text(str(session_record.sessionID), status=201)


@service.put("/registerSuite")
async def register_suite(request: Request) -> HTTPResponse:
    suite = RegisterSuite.model_validate(request.json)
    suite_record = await SuiteBase.create(**suite.model_dump())
    await suite_record.save()

    return text(str(suite_record.suiteID), status=201)


@service.put("/updateSuite", error_format="json")
async def updateSuite(request: Request) -> HTTPResponse:
    suite = MarkSuite.model_validate(request.json)

    suite_record = await SuiteBase.filter(suiteID=suite.suiteID).first()
    if not suite_record:
        logger.error("Was not able to found {} suite", str(suite.suiteID))
        return text(f"Suite {suite.suiteID} was not found", status=404)

    if suite_record.suiteType == SuiteType.SUITE:
        suite.standing = Status.YET_TO_CALCULATE
    else:
        note = {
            suite.standing.lower(): 1,
            "tests": 1,
        }
        await suite_record.update_from_dict(note)
        await suite_record.save()

    await suite_record.update_from_dict(suite.model_dump())
    await suite_record.save()

    if suite_record.suiteType == SuiteType.SUITE:
        await register_patch_suite(suite_record.suiteID, get_test_id())

    return text(
        f"Suite: {suite_record.title} - {suite_record.suiteID} was updated", status=201
    )


@service.put("/updateSession", error_format="json")
async def update_session(request: Request) -> HTTPResponse:
    session = MarkSession.model_validate(request.json)
    test_session = await SessionBase.filter(sessionID=session.sessionID).first()
    if not test_session:
        logger.error("Expected {} session was not found", str(session.sessionID))
        return text(f"Session {session.sessionID} was not found", status=404)

    await test_session.update_from_dict(session.model_dump())
    await test_session.save()
    return text(f"{session.sessionID} was updated", status=201)


@service.put("/addAttachmentsForEntities")
async def addAttachmentForEntity(request: Request) -> HTTPResponse:
    attachments = []
    note = []
    assertions = []

    for _ in request.json:
        try:
            if _["type"] == AttachmentType.ASSERT:
                value = _.get("value", dict())
                value["wait"] = value.get("wait", -1)
                value["interval"] = value.get("interval", -1)

            attachment = AddAttachmentForEntity.model_validate(_)
        except ValidationError as error:
            note.append(_.get("entityID", False))
            url = "/addAttachmentsForEntities"
            await attachWarn(extractPydanticErrors(url, _, error), url)
            continue

        match attachment.type:
            case AttachmentType.ASSERT:
                assertions.append(
                    await AssertBase(
                        **dict(
                            entity_id=attachment.entityID,
                            title=attachment.title,
                            message=attachment.value["message"],
                            passed=attachment.value["passed"],
                            interval=attachment.value["interval"],
                            wait=attachment.value["wait"],
                        )
                    )
                )

            case _:
                (
                    attachments.append(
                        await AttachmentBase(
                            **dict(
                                entity_id=attachment.entityID,
                                description=attachment.description,
                                type=attachment.type,
                                attachmentValue=dict(
                                    color=attachment.color,
                                    value=attachment.value,
                                    title=attachment.title,
                                ),
                            )
                        )
                    )
                )

    if attachments:
        await AttachmentBase.bulk_create(attachments)
    if assertions:
        await AssertBase.bulk_create(assertions)
    return text(
        "Attachments was added successfully", status=201 if not len(note) else 206
    )


@service.put("/currentRun")
async def update_run_config(request: Request) -> HTTPResponse:
    run_config = PydanticModalForTestRunConfigBase.model_validate(request.json)
    config = await TestConfigBase.create(
        test_id=get_test_id(),
        platform=run_config.platformName,
        framework=run_config.framework,
        maxInstances=run_config.maxInstances,
        fileRetries=run_config.fileRetries,
        avoidParentSuitesInCount=run_config.avoidParentSuitesInCount,
        bail=run_config.bail,
        tags=run_config.tags,
    )

    test = await config.test
    await test.update_from_dict(dict(exitCode=run_config.exitCode))
    await test.save()
    await config.save()

    return text("provided config was saved successfully.", status=200)


@service.put("/registerAWrittenAttachment", error_format="json")
async def saveImage(request: Request) -> HTTPResponse:
    attachment = WrittenAttachmentForEntity.model_validate(request.json)
    record = await StaticBase.create(
        entity_id=attachment.entityID,
        description=attachment.description,
        type=attachment.type,
    )
    file_name = f"{record.attachmentID}.{record.type.lower()}"
    await record.update_from_dict(
        dict(
            attachmentValue=dict(value=file_name, title=attachment.title),
        )
    )
    await record.save()
    # we can save the file in this request itself, but no. we let the framework's custom reporter cook.
    return text(
        str(attachment_folder(db_path()) / get_test_id() / file_name), status=201
    )
